/* 
**  mod_auth_openpgp.c -- Apache auth_openpgp module
**  Author: Arturo 'Buanzo' Busleiman <buanzo@buanzo.com.ar>
**  Based on ideas from Apache's mod_authz_host.c, mod_echo.c, mod_access.c
**  Lots of Googling and great feedback from modules-dev@apache.org made this possible.
**
**  This Apache Module implements OpenPGP operations over HTTP
**  as explained in the openpgp-extensions-for-http draft by Arturo Busleiman.
**
**  Licensed under the Apache License 2.0
*/

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include <apr_uri.h>
//#include "ap_config.h"

#include <gpgme.h>

/* BUANZONOTES: This comment contains useful code I haven't used yet
ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
        "Incomplete request_rec!") ;

*/
module AP_MODULE_DECLARE_DATA auth_openpgp_module;

static int auth_openpgp_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *main_s) {
	ap_add_version_component(p, "mod_auth_openpgp/0.2.1"); 
	return OK;
}

/* begin CONFIG section */
typedef struct AuthOpenPGPConfig {
	int bEnabled;
	char *KeyServer;
} AuthOpenPGPConfig;

static void *create_auth_openpgp_server_config(apr_pool_t *p, server_rec *s) {
AuthOpenPGPConfig *pConfig = apr_pcalloc(p, sizeof (AuthOpenPGPConfig));

pConfig->bEnabled = 0;
pConfig->KeyServer = (char*)apr_pstrdup(p,"hkp://subkeys.pgp.net");
return pConfig;
}

static const char *auth_openpgp_on(cmd_parms *cmd, void *dummy, int arg) {
AuthOpenPGPConfig *pConfig = ap_get_module_config(cmd->server->module_config,
                                               &auth_openpgp_module);
pConfig->bEnabled = arg;

return NULL;
}

static const char *auth_openpgp_set_keyserver(cmd_parms *cmd, void *dummy, const char *val) {
AuthOpenPGPConfig *pConfig = ap_get_module_config(cmd->server->module_config,
                                               &auth_openpgp_module);
pConfig->KeyServer = (char *)apr_pstrdup(cmd->pool,val);

return NULL;
}

static const command_rec auth_openpgp_cmds[] = {
	AP_INIT_FLAG("OpenPGPEngine", auth_openpgp_on, NULL, RSRC_CONF, "Enable OpenPGP Auth on this host"),
	AP_INIT_TAKE1("OpenPGPKeyServer",auth_openpgp_set_keyserver,NULL,RSRC_CONF,"Specify which keyserver to use for public key imports. Same syntax as gpg\'s --keyserver. Default: hkp://subkeys.pgp.net"),
	{ NULL }
};

/* end CONFIG section */

int
gpg_verification (char *message, char *signature, int msgSize, char **keyid,char **fpr, char **email, char **name, char **comment, request_rec * r) {
gpgme_ctx_t		ctx;
gpgme_error_t	err;
gpgme_data_t	sign, text;
gpgme_verify_result_t result;
gpgme_key_t		key;

int retval = 1;

if (msgSize < 0) return 1;

err = gpgme_new (&ctx);
if (msgSize == 0)
	err = gpgme_data_new_from_mem (&text, message, strlen(message), 0);
else
	err = gpgme_data_new_from_mem (&text, message, msgSize, 0);

err = gpgme_data_new_from_mem (&sign, signature, strlen (signature), 0);
err = gpgme_op_verify (ctx, sign, text, NULL);
result = gpgme_op_verify_result (ctx);

if (result->signatures->fpr != NULL) {
	/* copy fingerprint */
	*fpr = apr_pcalloc (r->pool, strlen(result->signatures->fpr) + 1);
	memcpy ((char *) *fpr, result->signatures->fpr, strlen(result->signatures->fpr));
} else
	fpr=NULL;

switch (gpg_err_code (result->signatures->status)) {
case GPG_ERR_NO_ERROR: {
	err = gpgme_get_key (ctx, result->signatures->fpr,  &key, 0);
	if (err==GPG_ERR_NO_ERROR) {
	if (key->subkeys->keyid != NULL) {
		*keyid = apr_pcalloc (r->pool, strlen(key->subkeys->keyid) + 1);
		memcpy ((char *) *keyid, key->subkeys->keyid, strlen(key->subkeys->keyid));
	}
	if (key->uids->email != NULL) {
		*email = apr_pcalloc (r->pool, strlen(key->uids->email) + 1);
		memcpy ((char *) *email, key->uids->email, strlen(key->uids->email));
	}
	if (key->uids->name != NULL) {
		*name = apr_pcalloc (r->pool, strlen(key->uids->name) + 1);
		memcpy ((char *) *name, key->uids->name, strlen(key->uids->name));
	}
	if (key->uids->comment != NULL) {
		*comment = apr_pcalloc (r->pool, strlen(key->uids->comment) + 1);
		memcpy ((char *) *comment, key->uids->comment, strlen(key->uids->email));
	}
	}
	retval = 0;
	break;
}
/*
case GPG_ERR_BAD_SIGNATURE:
printf("GPGME_SIG_STAT_BAD\n");
//*r_stat = GPGME_SIG_STAT_BAD;
break;

case GPG_ERR_NO_PUBKEY:
printf("GPGME_SIG_STAT_NOKEY\n");
//*r_stat = GPGME_SIG_STAT_NOKEY;
break;

case GPG_ERR_NO_DATA:
printf("GPGME_SIG_STAT_NOSIG\n");
//*r_stat = GPGME_SIG_STAT_NOSIG;
break;

case GPG_ERR_SIG_EXPIRED:
printf("GPGME_SIG_STAT_GOOD_EXP\n");
//*r_stat = GPGME_SIG_STAT_GOOD_EXP;
break;

case GPG_ERR_KEY_EXPIRED:
printf("GPGME_SIG_STAT_GOOD_EXPKEY\n");
//*r_stat = GPGME_SIG_STAT_GOOD_EXPKEY;
break;
*/

default: {
	retval = 1;
	break;
}
}

gpgme_data_release (sign);
gpgme_data_release (text);
gpgme_release (ctx);
return (retval);
}

static int util_read (request_rec * r, char **rbuf) {
int rc=OK;

if ((rc = ap_setup_client_block (r, REQUEST_CHUNKED_ERROR)) != OK) return rc;

if (ap_should_client_block (r)) {
	char argsbuffer[HUGE_STRING_LEN];
	int rsize, len_read, rpos = 0;
	long length = r->remaining;
	
	*rbuf = apr_pcalloc (r->pool, length + 1);

	while ((len_read = ap_get_client_block (r, argsbuffer, sizeof (argsbuffer))) > 0) {
		if ((rpos + len_read) > length)
			rsize = length - rpos;
		else
			rsize = len_read;
		memcpy ((char *) *rbuf + rpos, argsbuffer, rsize);
		rpos += rsize;
	}

}
return rc;
}

static int
openpgp_discovery_handler (request_rec * r) {

AuthOpenPGPConfig *pConfig = ap_get_module_config(r->server->module_config, &auth_openpgp_module);

if (!pConfig->bEnabled || !r->handler || strcasecmp(r->handler, "mao_Discovery") != 0) {
	return DECLINED;
}

if (r->method_number != M_GET) {
	return HTTP_METHOD_NOT_ALLOWED;
}

apr_table_set(r->headers_out,"X-OpenPGP-Agent","mod_auth_openpgp/0.2.1");
ap_set_content_type(r, "text/plain");
ap_rputs("HTTP-OpenPGP-Supported",r);

return OK;
}

static int
import_key_handler (request_rec * r) {
const char *fpr = NULL;
AuthOpenPGPConfig *pConfig = ap_get_module_config(r->server->module_config, &auth_openpgp_module);

	if (!r->handler || strcasecmp(r->handler, "mao_ImportKey") != 0) {
		return DECLINED;
	}

	if (r->method_number != M_GET) {
		return HTTP_METHOD_NOT_ALLOWED;
	}


	fpr = apr_table_get (r->headers_in, "X-Auth-OpenPGP-Fingerprint");
	//apr_table_set(r->headers_out,"X-OpenPGP-ImportKeyUrl",);
	ap_set_content_type(r, "text/html");
	ap_rputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\">\n",r);

	ap_rputs("<HTML><HEAD><title>mod_auth_openpgp Import Key Handler</title></HEAD><BODY><P>Hello, this is mod_auth_openpgp Import Key Handler.</P>", r);
	ap_rputs("<P>LOC = ",r);
	ap_rputs((const char*)apr_psprintf(r->pool,"http://%s%s",r->hostname,r->uri),r);
	ap_rputs("</P>",r);
	ap_rputs("<P> Key Import Server = ",r);
	ap_rputs(pConfig->KeyServer,r);
	ap_rputs("</P>",r);
	if (fpr != NULL) {
		ap_rputs("<P>Importing OpenPGP Key of Fingerprint = ", r);
		ap_rputs(fpr, r);
		ap_rputs(".</P>",r);
//		r->args = apr_psprintf(r->pool,"--keyserver hkp://subkeys.pgp.net --recv-keys %s",fpr);
//		apr_call_exec(r, "/usr/bin/gpg", NULL, 0);
	}
	ap_rputs("</BODY></HTML>", r);
	

return OK;
}

static int
auth_openpgp_handler (request_rec * r)
{
const char *openpgp_type = NULL, *openpgp_fields = NULL, *openpgp_sig = NULL,
*openpgp_algo = NULL, *openpgp_version = NULL, *openpgp_agent = NULL;
char *signature = NULL, *message = NULL, *keyid=NULL, *fingerprint=NULL, *email=NULL, *name=NULL, *comment=NULL;
char argsbuffer[HUGE_STRING_LEN];
int msgsize;
int i=-1;

AuthOpenPGPConfig *pConfig = ap_get_module_config(r->server->module_config, &auth_openpgp_module);

if (!pConfig->bEnabled) {
	return DECLINED;
}

/* remove spoofed headers. If module is not enabled for vhost, then if these headers
already exist in the request, they shouldn't be removed, but if it IS enabled... is considered spoofing. */

apr_table_unset (r->headers_in, "X-Auth-OpenPGP");
apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Fingerprint");
apr_table_unset (r->headers_in, "X-Auth-OpenPGP-KeyID");
apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Email");
apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Name");
apr_table_unset (r->headers_in, "X-Auth-OpenPGP-Comment");

/* read X-OpenPGP headers */
openpgp_type = apr_table_get (r->headers_in, "X-OpenPGP-Type");	/* As of Enigform 0.7.9, only S[ign] */
openpgp_fields = apr_table_get (r->headers_in, "X-OpenPGP-Sig-Fields");	/* as of Enigform 0.7.9, only 'body' */
openpgp_sig = apr_table_get (r->headers_in, "X-OpenPGP-Sig");
openpgp_algo = apr_table_get (r->headers_in, "X-OpenPGP-Digest-Algo");	/* not needed with gpgme */
openpgp_version = apr_table_get (r->headers_in, "X-OpenPGP-Version");
openpgp_agent = apr_table_get (r->headers_in, "X-OpenPGP-Agent");

/* Only accept POST and GET HTTP Methods [for now!] */
if (r->method_number != M_POST && r->method_number != M_GET) return HTTP_METHOD_NOT_ALLOWED;

if (openpgp_sig == NULL || openpgp_type==NULL || openpgp_type[0] != 'S') {
	//TODO: config item. Should HTTP_FORBIDDEN or OK for this server/location/dir/vhost/etc
	//TODO: most flexible behaviour until configuration parameter exist is set X-auth-openpgp to false and return OK
	apr_table_set(r->headers_in, "X-Auth-OpenPGP", "false");
	return OK;
}


/* Format Signature */
signature = (char *) apr_pstrcat (r->pool, "-----BEGIN PGP SIGNATURE-----\nVersion: ", openpgp_version, "\n\n", openpgp_sig, "\n-----END PGP SIGNATURE------\n", NULL);

/* Load payload into message var, depending on method_number (GET, POST, etc) */

switch (r->method_number) {
	case M_GET: {
		message = (char *) apr_pstrcat (r->pool, r->args, NULL);
		msgsize=0; // so gpg_verification uses strlen() without -1
		break;
	}
	case M_POST: {
		if (util_read(r,&message)!=OK)
			return HTTP_METHOD_NOT_ALLOWED;
		
		if (ap_strstr((char*)apr_table_get(r->headers_in,"Content-Type"),"boundary="))
			msgsize = atoi(apr_table_get(r->headers_in, "Content-Length"))-1;
		else
			msgsize = atoi(apr_table_get(r->headers_in, "Content-Length"));
		break;
	}
}

/* verify the message. */
i = gpg_verification (message, signature,msgsize,&keyid,&fingerprint,&email,&name,&comment,r);
if (fingerprint != NULL) apr_table_set (r->headers_in,"X-Auth-OpenPGP-Fingerprint",fingerprint);

if (i == 0) {
	apr_table_set (r->headers_in, "X-Auth-OpenPGP", "true");
	if (keyid != NULL)
		apr_table_set (r->headers_in,"X-Auth-OpenPGP-KeyID",keyid);
	if (email!=NULL)
		apr_table_set (r->headers_in,"X-Auth-OpenPGP-Email",email);
	if (name!=NULL)
		apr_table_set (r->headers_in,"X-Auth-OpenPGP-Name",name);
	if (comment!=NULL)
		apr_table_set (r->headers_in,"X-Auth-OpenPGP-Comment",comment);
	return OK;
} else {
	apr_table_set(r->headers_in, "X-Auth-OpenPGP", "false");
	return OK;
//	return HTTP_FORBIDDEN;
}

} /* end auth_openpgp_handler() */

static void auth_openpgp_register_hooks (apr_pool_t * p) {
	ap_hook_post_config(auth_openpgp_init, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_post_read_request (auth_openpgp_handler, NULL, NULL, APR_HOOK_FIRST);
	ap_hook_handler(import_key_handler, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_handler(openpgp_discovery_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

/* Dispatch list for API hooks */
module AP_MODULE_DECLARE_DATA auth_openpgp_module = {
	STANDARD20_MODULE_STUFF,
	NULL, /* create per-dir    config structures */
	NULL, /* merge  per-dir    config structures */
	create_auth_openpgp_server_config, /* create_auth_openpgp_server_config create per-server config structures */
	NULL, /* merge  per-server config structures */
	auth_openpgp_cmds, /* table of config file commands */
	auth_openpgp_register_hooks /* register hooks */
};
